#!/usr/bin/env python3

# The code in this file is released under the Creative Commons CC0 license,
# i.e., for unrestricted use and without any warranty,
# see the file license_CC0.txt or https://creativecommons.org/publicdomain/zero/1.0/

# It requires part of the Lahman Baseball Database (2014 Version) in the subfolder data: Batting.csv, Fielding.csv, Pitching.csv, and People.csv.
# This data is available from https://github.com/chadwickbureau/baseballdatabank.git
# and subject to Creative Commons Attribution-ShareAlike 3.0 Unported License. For details see: http://creativecommons.org/licenses/by-sa/3.0/

from itertools import product
import numpy as np
import pandas as pd
from scipy.stats import mannwhitneyu
from tqdm import tqdm


# US and Canadian state abbreviations as used in the data
states = ('VA', 'ID', 'UT', 'MT', 'MB', 'TN', 'WV', 'NE', 'CA', 'IN', 'LA', 'WA', 'NY', 'NM', 'ND', 'OH', 'FL', 'NJ', 'MD', 'NV', 'SK', 'AZ', 'NC', 'AB', 'WY', 'IL', 'CO', 'SC', 'DC', 'NS', 'MI', 'MN', 'PA', 'MA', 'KY', 'AK', 'CT', 'QC', 'AR', 'GA', 'MO', 'ME', 'KS', 'TX', 'HI', 'OK', 'DE', 'VT', 'MS', 'IA', 'OR', 'SD', 'BC', 'NH', 'ON', 'RI', 'NB', 'WI', 'AL', 'PE')
# columns in the datasets
batting_cols = ('G', 'AB', 'R', 'H', '2B', '3B', 'HR', 'RBI', 'SB', 'CS', 'BB', 'SO', 'IBB', 'HBP', 'SH', 'SF', 'GIDP')
fielding_cols = ('G', 'GS', 'InnOuts', 'PO', 'A', 'E', 'DP', 'PB', 'WP', 'SB', 'CS', 'ZR')
pitching_cols = ('W', 'L', 'G', 'GS', 'CG', 'SHO', 'SV', 'IPouts', 'H', 'ER', 'HR', 'BB', 'SO', 'BAOpp', 'ERA', 'IBB', 'WP', 'HBP', 'BK', 'BFP', 'GF', 'R', 'SH', 'SF', 'GIDP')


def pValueOfStatTest(values_a, values_b):
    return mannwhitneyu(values_a, values_b, alternative='two-sided')[1]


def loadData():
    def loadMergeAndFilter(people, filename):
        data_frame = pd.read_csv(filename)
        data_frame_ = pd.merge(people, data_frame, on='playerID')
        return data_frame_.loc[data_frame_ ['birthCountry'].isin(('USA', 'CAN'))]

    people = pd.read_csv('data/People.csv')

    return( loadMergeAndFilter(people, 'data/Batting.csv'),
            loadMergeAndFilter(people, 'data/Fielding.csv'),
            loadMergeAndFilter(people, 'data/Pitching.csv') )


def pHack(data, column_names, select_column, separate_values, data_identifier):
    def summarizeData(raw_data):
        grouped_data_by_player = {}
        for player_id, selector, value in raw_data:
            if player_id not in grouped_data_by_player:
                grouped_data_by_player[player_id] = []
            grouped_data_by_player[player_id].append((selector, value))
        ret = []
        for selector_value_list in grouped_data_by_player.values():
            # selector_value_list[j][0] should, by construction, be the same for all j
            ret.append((selector_value_list[0][0], np.sum([i[1] for i in selector_value_list])))
        return ret

    def splitInGroups(data_by_col, separate_values):
        data_in_group = {}
        for sep_val in separate_values:
            data_in_group[sep_val] = []

        for row in data_by_col:
            group_value = row[0]
            row_value = row[1]
            use_value = None
            try:
                use_value = not np.isnan(group_value)
            except:
                use_value = (group_value != '')
            if use_value and not np.isnan(row_value) and row_value != 0:
                # filter unavailable values (nan) and probably not-applicable (0) values
                data_in_group[group_value].append(row_value)
        return data_in_group

    def computeAndAppendPValue(data_1, data_2, data_identifier, results):
        if len(data_1) > 20 and len(data_2) > 20:
            # statistical test requires at least 20 samples
            p_value = pValueOfStatTest(data_1, data_2)
            results.append((p_value,
                            data_identifier,
                            len(data_1), np.median(data_1), (np.percentile(data_1, 75) - np.percentile(data_1, 25)),
                            len(data_2), np.median(data_2), (np.percentile(data_2, 75) - np.percentile(data_2, 25))))
        return results

    results = []
    for col in tqdm(column_names):
        data_by_col = summarizeData(data.as_matrix(columns=('playerID', select_column, col)))
        data_in_group = splitInGroups(data_by_col, separate_values)

        for val1, val2 in product(separate_values, separate_values):
            if val1 < val2:
                # exploit symmetry and don't compare to self
                results = computeAndAppendPValue(data_in_group[val1], data_in_group[val2],
                                                 data_identifier + col + ' ' + str(val1) + '/' + str(val2), results)

    return results


def doPHacking(batting, fielding, pitching):
    def verifyResults(results):
        # this L2 difference should be small, the p values should approximately follow a uniform distribution
        p_values_data = np.array([i[0] for i in results])
        uniform_values_same_length = np.linspace(0, 1.0, len(p_values_data))
        return '# ' + str(np.linalg.norm(p_values_data - uniform_values_same_length, ord=2)/len(p_values_data)) + '\n'

    for col_name, groups, filename in (('birthDay', range(1, 32), 'byBirthDay.out'),
                                       ('birthMonth', range(1, 13), 'byBirthMonth.out'),
                                       ('birthState', states, 'byBirthState.out'),
                                       ('deathDay', range(1, 32), 'byDeathDay.out'),
                                       ('deathMonth', range(1, 13), 'byDeathMonth.out')):
        results = []
        for data, cols, col_id in ((batting, batting_cols, 'b-'), (fielding, fielding_cols, 'f-'), (pitching, pitching_cols, 'p-')):
            results += pHack(data, cols, col_name, groups, col_id)
        with open(filename, 'w') as outfile:
            outfile.write(verifyResults(results))

            outfile.write('# p-value data_column groups_compared_a/b size_a median_a IQR_a size_b median_b IQR_b\n')
            for i in sorted(results):
                outfile.write(str(i)+'\n')


if __name__ == '__main__':
    batting, fielding, pitching = loadData()
    doPHacking(batting, fielding, pitching)
